@@ -20,14 +20,15 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
 from message import Message
 from common import IntDict, Debug
+from tile import Tile
 from meld import elementKey
 
-class AIDefault:
+class AIDefault(object):
     """all AI code should go in here"""
 
     groupPrefs = {'s':0, 'b':0, 'c':0, 'w':4, 'd':7}
 
-    # pylint: disable=R0201
+    # pylint: disable=no-self-use
     # we could solve this by moving those filters into DiscardCandidates
     # but that would make it more complicated to define alternative AIs
 
@@ -40,7 +41,7 @@ class AIDefault:
 
     @staticmethod
     def weighSameColors(dummyAiInstance, candidates):
-        """weigh tiles of same color against each other"""
+        """weigh tiles of same group against each other"""
         for candidate in candidates:
             if candidate.group in 'sbc':
                 if candidate.prev.occurrence:
@@ -58,14 +59,16 @@ class AIDefault:
         return candidates
 
     def selectDiscard(self, hand):
-        # pylint: disable=R0912, R0915
+        # pylint: disable=too-many-branches, R0915
         # disable warning about too many branches
         """returns exactly one tile for discard.
         Much of this is just trial and success - trying to get as much AI
         as possible with limited computing resources, it stands on
         no theoretical basis"""
         candidates = DiscardCandidates(self.client.game, hand)
-        return self.weighDiscardCandidates(candidates).best()
+        result = self.weighDiscardCandidates(candidates).best()
+        candidates.unlink()
+        return result
 
     def weighDiscardCandidates(self, candidates):
         """the standard"""
@@ -102,7 +105,7 @@ class AIDefault:
     @staticmethod
     def weighBasics(aiInstance, candidates):
         """basic things"""
-        # pylint: disable=R0912
+        # pylint: disable=too-many-branches
         # too many branches
         for candidate in candidates:
             keep = candidate.keep
@@ -152,14 +155,14 @@ class AIDefault:
             group = candidate.group
             groupCount = candidates.groupCounts[group]
             if group in 'sbc':
-                # count tiles with a different color:
+                # count tiles with a different group:
                 if groupCount == 1:
                     candidate.keep -= 2.013
                 else:
                     otherGC = sum(candidates.groupCounts[x] for x in 'sbc' if x != group)
                     if otherGC:
                         if groupCount > 8 or otherGC < 5:
-                            # do not go for color game if we already declared something in another color:
+                            # do not go for color game if we already declared something in another group:
                             if not any(candidates.declaredGroupCounts[x] for x in 'sbc' if x != group):
                                 candidate.keep += 20 // otherGC
             elif group == 'w' and groupCount > 8:
@@ -215,7 +218,7 @@ class AIDefault:
     def selectAnswer(self, answers):
         """this is where the robot AI should go.
         Returns answer and one parameter"""
-        # pylint: disable=R0912
+        # pylint: disable=too-many-branches
         # disable warning about too many branches
         answer = parameter = None
         tryAnswers = (x for x in [Message.MahJongg, Message.OriginalCall, Message.Kong,
@@ -225,7 +228,7 @@ class AIDefault:
         discard = self.client.game.lastDiscard
         if discard:
             for func in self.client.game.ruleset.filterFunctions('claimness'):
-                claimness += func.claimness(hand, discard.element)
+                claimness += func.claimness(hand, discard)
         for tryAnswer in tryAnswers:
             parameter = self.client.sayable[tryAnswer]
             if not parameter:
@@ -261,8 +264,8 @@ class AIDefault:
                 if not myself.hasConcealedTiles(chow):
                     # do not dissolve an existing chow
                     belongsToPair = False
-                    for tileName in chow:
-                        if myself.concealedTileNames.count(tileName) == 2:
+                    for tile in chow:
+                        if myself.concealedTileNames.count(tile) == 2:
                             belongsToPair = True
                             break
                     if not belongsToPair:
@@ -319,9 +322,10 @@ class AIDefault:
 
 class TileAI(object):
     """holds a few AI related tile properties"""
-    # pylint: disable=R0902
+    # pylint: disable=too-many-instance-attributes
     # we do want that many instance attributes
     def __init__(self, candidates, name):
+        assert isinstance(name, Tile), name
         self.name = name
         self.group, self.value = name[:2]
         if self.value in '123456789bgreswn' and len(name) == 2:
@@ -341,11 +345,13 @@ class TileAI(object):
         self.prev2 = None
         self.next2 = None
 
-
     def __str__(self):
         dang = ' dang:%d' % self.dangerous if self.dangerous else ''
         return '%s:=%s%s' % (self.name, self.keep, dang)
 
+    def __repr__(self):
+        return 'TileAI(%s)' % str(self)
+
 class DiscardCandidates(list):
     """a list of TileAI objects. This class should only hold
     AI neutral methods"""
@@ -353,12 +359,12 @@ class DiscardCandidates(list):
         list.__init__(self)
         self.game = game
         self.hand = hand
-        self.hiddenTiles = list(x.lower() for x in hand.inHand)
+        self.hiddenTiles = list(x.lower() for x in hand.tilesInHand)
         self.groupCounts = IntDict() # counts for tile groups (sbcdw), exposed and concealed
         for tile in self.hiddenTiles:
             self.groupCounts[tile[0]] += 1
         self.declaredGroupCounts = IntDict()
-        for tile in sum((x.pairs.lower() for x in hand.declaredMelds), []):
+        for tile in sum((x.lower() for x in hand.declaredMelds), []):
             self.groupCounts[tile[0]] += 1
             self.declaredGroupCounts[tile[0]] += 1
         self.extend(list(TileAI(self, x) for x in sorted(set(self.hiddenTiles), key=elementKey)))
@@ -386,15 +392,23 @@ class DiscardCandidates(list):
             if this.group in 'sbc':
                 # we want every tile to have prev/prev2/next/next2
                 # the names do not matter, just occurrence, available etc
-                thisValue = this.value
+                thisValue = ord(this.value)
                 if not this.prev:
-                    this.prev = TileAI(self, this.group +  str(int(thisValue)-1))
+                    this.prev = TileAI(self, Tile(this.group +  chr(thisValue-1)))
                 if not this.prev2:
-                    this.prev2 = TileAI(self, this.group +  str(int(thisValue)-2))
+                    this.prev2 = TileAI(self, Tile(this.group +  chr(thisValue-2)))
                 if not this.next:
-                    this.next = TileAI(self, this.group +  str(int(thisValue)+1))
+                    this.next = TileAI(self, Tile(this.group +  chr(thisValue+1)))
                 if not this.next2:
-                    this.next2 = TileAI(self, this.group +  str(int(thisValue)+2))
+                    this.next2 = TileAI(self, Tile(this.group +  chr(thisValue+2)))
+
+    def unlink(self):
+        """remove links between elements. This helps garbage collection."""
+        for this in self:
+            this.prev = None
+            this.next = None
+            this.prev2 = None
+            this.next2 = None
 
     def best(self):
         """returns the candidate with the lowest value"""
